from abc import ABC, abstractmethod
from typing import Generic

from ..types import FIG_T, PLOT_DATA_T


class BasePlotter(Generic[PLOT_DATA_T, FIG_T], ABC):
    @abstractmethod
    def render(self, plot_data: PLOT_DATA_T) -> FIG_T:
        """Render a Plotly figure from extracted data."""
        pass



from __future__ import annotations

from typing import ClassVar

import polars as pl
import polars.selectors as cs

from .base import BaseExtractor
from .meds_pl_mixin import MEDSColumns


class CodeFrequency(BaseExtractor[pl.DataFrame, pl.DataFrame], MEDSColumns):
    """An extractor that counts the frequency of each code in a given dataset."""

    as_proportions: bool
    sort_by: str
    code_counts: pl.DataFrame | None
    n_subjects: int | None
    n_events: int | None
    n_measurements: int | None
    n_codes: int | None

    CODE_AGG_EXPRS: ClassVar[dict[str, pl.Expr]] = {
        "code/n_occurrences": pl.len().alias("code/n_occurrences"),
        "code/n_subjects": MEDSColumns.SUBJECT_ID.n_unique().alias("code/n_subjects"),
        "code/n_events": MEDSColumns.TIME.n_unique().alias("code/n_events"),
    }

    GLOBAL_AGG_EXPRS: ClassVar[dict[str, pl.Expr]] = {
        "n_subjects": MEDSColumns.SUBJECT_ID.n_unique().alias("n_subjects"),
        "n_events": MEDSColumns.TIME.n_unique().alias("n_events"),
        "n_measurements": pl.len().alias("n_measurements"),
    }

    def __init__(self, as_proportions: bool = True, sort_by: str = "n_occurrences"):
        super().__init__()
        self.as_proportions = as_proportions
        self.sort_by = sort_by
        self.code_counts = None
        self.n_subjects = None
        self.n_events = None
        self.n_measurements = None

    def _extract(self, shard: pl.DataFrame, code_metadata: pl.DataFrame | None = None):
        self.code_counts = shard.group_by("code").agg(**self.CODE_AGG_EXPRS)
        global_aggs = shard.select(**self.GLOBAL_AGG_EXPRS)

        for agg_name in self.GLOBAL_AGG_EXPRS:
            setattr(self, agg_name, global_aggs[agg_name].item())

    def _merge(self, other: CodeFrequency, code_metadata: pl.DataFrame | None = None):
        for agg_name in self.GLOBAL_AGG_EXPRS:
            setattr(self, agg_name, getattr(self, agg_name) + getattr(other, agg_name))

        self.code_counts = (
            pl.concat([self.code_counts, other.code_counts], how="vertical")
            .group_by(self.CODE)
            .agg(cs.starts_with("code/").sum().name.keep())
        )

    @property
    def proportion_exprs(self) -> dict[str, pl.Expr]:
        return {
            "code/n_occurrences": (pl.col("code/n_occurrences") / self.n_measurements),
            "code/n_subjects": (pl.col("code/n_subjects") / self.n_subjects),
            "code/n_events": (pl.col("code/n_events") / self.n_events),
        }

    @property
    def plot_data(self) -> pl.DataFrame:
        if self.as_proportions:
            df = self.code_counts.with_columns(**self.proportion_exprs)
        else:
            df = self.code_counts

        return df.select(
            self.CODE,
            cs.starts_with("code/").name.map(lambda x: x.removeprefix("code/")),
        ).sort(by=self.sort_by, descending=True)



from typing import Generic

from .extractors import BaseExtractor
from .plotters import BasePlotter
from .types import DF_T, FIG_T, PLOT_DATA_T


class Visualization(Generic[DF_T, PLOT_DATA_T, FIG_T]):
    def __init__(self, extractor: BaseExtractor[DF_T, PLOT_DATA_T], plotter: BasePlotter[PLOT_DATA_T, FIG_T]):
        self.extractor = extractor
        self.plotter = plotter

        if self.extractor.is_fit:
            raise ValueError("Extractor has already been fit. Please create a new instance.")

    def render(self, shards: list[DF_T]) -> FIG_T:
        self.extractor = self.extractor.fit(shards)
        return self.plotter.render(self.extractor.plot_data)



import plotly.graph_objects as go
import polars as pl

from .base import BasePlotter


class Bar(BasePlotter[pl.DataFrame, go.Figure]):
    top_k: int
    x_col: str | None
    y_cols: list[str] | None

    def __init__(self, top_k: int = 20, x_col: str | None = None, y_cols: list[str] | None = None):
        self.top_k = top_k
        self.x_col = x_col
        self.y_cols = y_cols

    def render(self, plot_data: pl.DataFrame) -> go.Figure:
        # Default to first column for X and all others for Y
        columns = plot_data.columns
        if not columns:
            raise ValueError("Empty dataframe passed to BarPlotter")

        x_col = self.x_col or columns[0]
        y_cols = self.y_cols or [c for c in columns if c != x_col]

        df = plot_data.head(self.top_k)

        fig = go.Figure()
        for y_col in y_cols:
            fig.add_trace(go.Bar(x=df[x_col].to_list(), y=df[y_col].to_list(), name=y_col))

        fig.update_layout(
            title=f"Bar Plot of {', '.join(y_cols)} by {x_col}",
            xaxis_title=x_col,
            yaxis_title="Values",
            barmode="group",
        )
        return fig



"""Test set-up and fixtures code."""

from datetime import datetime
from typing import Any

import polars as pl
import pytest


@pytest.fixture(autouse=True)
def __MEDS_visualizations_setup_doctest_namespace(doctest_namespace: dict[str, Any]) -> None:
    doctest_namespace.update(
        {
            "pl": pl,
            "datetime": datetime,
        }
    )



import plotly.graph_objects as go
import polars as pl

from .base import BasePlotter


class DensityHeatmap(BasePlotter[pl.DataFrame, go.Figure]):
    def __init__(self, x: str, y: str, z: str = "n_measurements"):
        self.x = x
        self.y = y
        self.z = z

    def render(self, plot_data: pl.DataFrame) -> go.Figure:
        return go.Figure(
            data=go.Histogram2d(
                x=plot_data[self.x].to_list(),
                y=plot_data[self.y].to_list(),
                z=plot_data[self.z].to_list(),
                coloraxis="coloraxis",
            )
        ).update_layout(coloraxis={"colorscale": "Viridis"}, title="2D Measurement Density")



import copy
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from ..types import DF_T, PLOT_DATA_T

EXTRACTOR_T = TypeVar("EXTRACTOR_T", bound="BaseExtractor")


class BaseExtractor(Generic[DF_T, PLOT_DATA_T], ABC):
    """A base class for data extractors used for visualizing MEDS datasets.

    A user must overwrite three methods to leverage the extractor:
      - `_extract`: Extract relevant data from a shard of data.
      - `_merge`: Merge outputs from multiple extractors into a unified representation.
    """

    is_fit: bool

    def __init__(self):
        self.is_fit = False

    @abstractmethod
    def _extract(self, shard: DF_T, code_metadata: DF_T | None = None):
        """Extract relevant data from the full dataset."""
        pass

    def _fit(self, shard: DF_T, code_metadata: DF_T | None = None) -> EXTRACTOR_T:
        self._extract(shard)
        self.is_fit = True
        return self

    def fit(self, data: DF_T | list[DF_T], code_metadata: DF_T | None = None) -> EXTRACTOR_T:
        """Fit the extractor to the data, fitting all internal parameters needed across all shards."""

        if self.is_fit:
            raise ValueError("Extractor has already been fit. Please create a new instance.")

        if not isinstance(data, list):
            data = [data]

        if len(data) == 0:
            raise ValueError("No data provided to fit the extractor.")

        extracted = [copy.deepcopy(self)._fit(shard, code_metadata) for shard in data[1:]]

        self._fit(data[0], code_metadata)
        self.merge(*extracted, code_metadata=code_metadata)

        return self

    @abstractmethod
    def _merge(self, other: EXTRACTOR_T, code_metadata: DF_T | None = None):
        pass

    def merge(self, *others: EXTRACTOR_T, code_metadata: DF_T | None = None) -> EXTRACTOR_T:
        """Merge outputs from multiple shards into a final representation."""

        if len(others) == 0:
            return self

        for other in others:
            if not isinstance(other, self.__class__):
                raise TypeError(f"Cannot merge {self.__class__.__name__} with {other.__class__.__name__}")
            if not other.is_fit:
                raise RuntimeError("Cannot merge unfit extractors.")
            self._merge(other, code_metadata=code_metadata)

        return self

    @property
    @abstractmethod
    def plot_data(self) -> PLOT_DATA_T:
        """Return the data to be plotted."""
        pass



from typing import ClassVar

import polars as pl
from meds import DataSchema


class MEDSColumns:
    """A Mixin class to add polars column expressions for the MEDS schema as class variables."""

    CODE: ClassVar[pl.Expr] = pl.col(DataSchema.code_name)
    SUBJECT_ID: ClassVar[pl.Expr] = pl.col(DataSchema.subject_id_name)
    TIME: ClassVar[pl.Expr] = pl.col(DataSchema.time_name)



from __future__ import annotations

import polars as pl

from .base import BaseExtractor
from .meds_pl_mixin import MEDSColumns


class ActiveSubjectsOverTime(BaseExtractor[pl.DataFrame, pl.DataFrame], MEDSColumns):
    """Estimates active subject count over a uniform time grid based on start/end times.

    TODO(mmd): Exclude birth events?

    Args:
        time_resolution: Either a polars duration string (e.g., '1d', '6h', '10m') to produce a uniform time
            grid at that frequency, the string "full" to use all unique time bounds (to see full granularity
            increases), or an integer as the number of unique time points to use in the grid (which will be
            chosen randomly from the unique time boundS).

    Examples:
        >>> shard = pl.DataFrame({
        ...     "subject_id": [
        ...         1, 1, 1, 1, 1, 1,
        ...         2, 2, 2, 2,
        ...     ],
        ...     "time": [
        ...         datetime(2020, 1, 1),
        ...         datetime(2020, 1, 1),
        ...         datetime(2020, 1, 2),
        ...         datetime(2020, 1, 3),
        ...         datetime(2020, 1, 5),
        ...         datetime(2020, 1, 8),
        ...         datetime(2020, 1, 4),
        ...         datetime(2020, 1, 4),
        ...         datetime(2020, 1, 4),
        ...         datetime(2020, 1, 10),
        ...     ],
        ... })
        >>> extractor = ActiveSubjectsOverTime(time_resolution="full")
        >>> _ = extractor.fit(shard)
        >>> extractor.plot_data
        shape: (4, 2)
        ┌─────────────────────┬─────────────────┐
        │ Time                ┆ Active Subjects │
        │ ---                 ┆ ---             │
        │ datetime[μs]        ┆ i64             │
        ╞═════════════════════╪═════════════════╡
        │ 2020-01-01 00:00:00 ┆ 1               │
        │ 2020-01-04 00:00:00 ┆ 2               │
        │ 2020-01-08 00:00:00 ┆ 1               │
        │ 2020-01-10 00:00:00 ┆ 0               │
        └─────────────────────┴─────────────────┘
    """

    time_resolution: str | int
    time_bounds: pl.DataFrame | None
    _seed: int

    def __init__(self, time_resolution: str | int = "1y", seed: int = 0):
        super().__init__()
        self.time_resolution = time_resolution
        self.time_bounds = None
        self._seed = seed

    def _extract(self, shard: pl.DataFrame):
        self.time_bounds = (
            shard.group_by(self.SUBJECT_ID)
            .agg(self.TIME.min().alias("min_time"), self.TIME.max().alias("max_time"))
            .select("min_time", "max_time")
        )

    def _merge(self, other: ActiveSubjectsOverTime, code_metadata: pl.DataFrame | None = None):
        self.time_bounds = pl.concat([self.time_bounds, other.time_bounds], how="vertical")

    @property
    def all_time_bounds(self) -> pl.DataFrame:
        """Returns the time bounds for all subjects in the dataset."""
        if self.time_bounds is None:
            raise ValueError("Time bounds have not been computed yet.")
        return pl.concat([self.time_bounds["min_time"], self.time_bounds["max_time"]]).unique()

    @property
    def min_time(self) -> pl.Series:
        """Returns the minimum time bound across all subjects."""
        if self.time_bounds is None:
            raise ValueError("Time bounds have not been computed yet.")
        return self.time_bounds["min_time"].min()

    @property
    def max_time(self) -> pl.Series:
        """Returns the maximum time bound across all subjects."""
        if self.time_bounds is None:
            raise ValueError("Time bounds have not been computed yet.")
        return self.time_bounds["max_time"].max()

    @property
    def plot_data(self) -> pl.DataFrame:
        match self.time_resolution:
            case "full":
                # Use all unique time bounds
                grid = self.all_time_bounds.sort()
            case int() as n_samples:
                # Randomly sample unique time bounds
                self._time_grid = self.all_time_bounds.sample(n=n_samples, random_state=self._seed)
            case str():
                # Use polars duration string to create a uniform time grid
                if not pl.duration(self.time_resolution).is_valid():
                    raise ValueError(f"Invalid time resolution: {self.time_resolution}")

                grid = pl.datetime_range(
                    low=self.min_time, high=self.max_time, interval=self.time_resolution, eager=True
                )

        # For each grid time t, compute active subjects
        active_counts = [
            ((self.time_bounds["min_time"] <= t) & (self.time_bounds["max_time"] > t)).sum() for t in grid
        ]

        return pl.DataFrame(
            {
                "Time": grid,
                "Active Subjects": active_counts,
            }
        )



import plotly.graph_objects as go
import polars as pl

from .base import BasePlotter


class Line(BasePlotter[pl.DataFrame, go.Figure]):
    x: str
    y: str

    def __init__(self, x: str, y: str):
        self.x = x
        self.y = y

    def render(self, plot_data: pl.DataFrame) -> go.Figure:
        return go.Figure(
            go.Scatter(x=plot_data[self.x].to_list(), y=plot_data[self.y].to_list(), mode="lines")
        ).update_layout(title=f"{self.y} over {self.x}", xaxis_title=self.x, yaxis_title=self.y)



from __future__ import annotations

from typing import TYPE_CHECKING

import polars as pl

from .base import BaseExtractor
from .meds_pl_mixin import MEDSColumns

if TYPE_CHECKING:
    from datetime import datetime


class MeasurementsInDuration(BaseExtractor[pl.DataFrame, pl.DataFrame], MEDSColumns):
    """Extracts data to visualize what duration of time is captured by a certain measurement window.

    This extractor is used to visualize the duration of time captured by a certain measurement window from a
    given index date across subjects.

    Args:
        index_date: The date from which to calculate the measurement window.
        min_offset: The minimum offset from the index date to include in the output. If None, no minimum
            offset is applied.
        max_offset: The maximum offset from the index date to include in the output. If None, no maximum
            offset is applied.
        filter_to_active_subjects: If True, only include subjects who have measurements on both sides of the
            index date are included. If False, all subjects are included.

    Attributes:
        index_date: The date from which to calculate the measurement window.
        min_offset: The minimum offset from the index date to include in the output.
        max_offset: The maximum offset from the index date to include in the output.
        filter_to_active_subjects: If True, only include subjects who have measurements on both sides of the
            index date are included.
        offset_deltas: A DataFrame containing the offsets and durations of measurements for each subject.

    Yields Plot Data:
        This extractor yields a DataFrame containing the offsets and durations of measurements for each
        subject-measurement. There will be two columns: "offset" and "delta". The "offset" column contains the
        number of measurements from the last measurement observed less than or equal to the index date, under
        the original shard ordering, and the "delta" column contains the duration from the time at that
        measurement to the index date (which may be zero). Both columns could be negative, zero, or positive.

    Examples:
        >>> shard = pl.DataFrame({
        ...     "subject_id": [
        ...         1, 1, 1, 1, 1, 1,
        ...         2, 2, 2, 2,
        ...     ],
        ...     "time": [
        ...         datetime(2020, 1, 1),
        ...         datetime(2020, 1, 1),
        ...         datetime(2020, 1, 2),
        ...         datetime(2020, 1, 3),
        ...         datetime(2020, 1, 5),
        ...         datetime(2020, 1, 8),
        ...         datetime(2020, 1, 4),
        ...         datetime(2020, 1, 4),
        ...         datetime(2020, 1, 4),
        ...         datetime(2020, 1, 10),
        ...     ],
        ... })
        >>> extractor = MeasurementsInDuration(index_date=datetime(2020, 1, 4))
        >>> _ = extractor.fit(shard)
        >>> extractor.plot_data
        shape: (10, 2)
        ┌────────┬──────────────┐
        │ offset ┆ delta        │
        │ ---    ┆ ---          │
        │ i32    ┆ duration[μs] │
        ╞════════╪══════════════╡
        │ -3     ┆ -3d          │
        │ -2     ┆ -3d          │
        │ -1     ┆ -2d          │
        │ 0      ┆ -1d          │
        │ 1      ┆ 1d           │
        │ 2      ┆ 4d           │
        │ -2     ┆ 0µs          │
        │ -1     ┆ 0µs          │
        │ 0      ┆ 0µs          │
        │ 1      ┆ 6d           │
        └────────┴──────────────┘
    """

    index_date: datetime
    min_offset: int | None
    max_offset: int | None
    filter_to_active_subjects: bool
    offset_deltas: pl.DataFrame | None

    def __init__(
        self,
        index_date: datetime,
        min_offset: int | None = None,
        max_offset: int | None = None,
        filter_to_active_subjects: bool = True,
    ):
        super().__init__()
        self.index_date = index_date
        self.min_offset = min_offset
        self.max_offset = max_offset
        self.filter_to_active_subjects = filter_to_active_subjects
        self.offset_deltas = None

    def _extract(self, shard: pl.DataFrame, code_metadata: pl.DataFrame | None = None):
        if self.filter_to_active_subjects:
            shard = shard.filter(
                (self.TIME.max().over(self.SUBJECT_ID) >= self.index_date)
                & (self.TIME.min().over(self.SUBJECT_ID) <= self.index_date)
            )

        row_idx = "__idx"

        shard = (
            shard.filter(self.TIME.is_not_null())
            .with_row_index(row_idx)
            .with_columns(pl.col(row_idx).cast(pl.Int32).name.keep())
        )

        with_deltas = shard.with_columns((self.TIME - self.index_date).alias("delta"))

        index_subj_row_idx = self.TIME.search_sorted(self.index_date, side="right").first() - 1

        index_date_indices = with_deltas.group_by(self.SUBJECT_ID).agg(
            pl.col(row_idx).get(index_subj_row_idx).alias("index_idx")
        )

        self.offset_deltas = with_deltas.join(
            index_date_indices, on=self.SUBJECT_ID, how="left", maintain_order="left"
        ).select(
            "delta",
            (pl.col(row_idx) - pl.col("index_idx")).alias("offset"),
        )

        if self.min_offset is not None:
            self.offset_deltas = self.offset_deltas.filter(pl.col("offset") >= self.min_offset)
        if self.max_offset is not None:
            self.offset_deltas = self.offset_deltas.filter(pl.col("offset") <= self.max_offset)

    def _merge(self, other: MeasurementsInDuration, code_metadata: pl.DataFrame | None = None):
        self.offset_deltas = pl.concat([self.offset_deltas, other.offset_deltas], how="vertical")

    @property
    def plot_data(self) -> pl.DataFrame:
        return self.offset_deltas.select("offset", "delta")
